/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//  Columbus High School Bottle Rocket Simulator                           //
//  Copyright (C) 1999-2000 Leo C. Singleton IV                            //
//    Internet: http://www.leosolutions.com/                               //
//    E-Mail: leo@leosolutions.com                                         //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

// Simulate.cpp

#include "stdafx.h"
#include "Simulate.h"

#include <math.h>
#include "Units.h"
#include "Variables.h"
#include "Graph.h"


CForces::CForces(CState *pState)
{
	m_pState = pState;
	m_pForce[0] = new CArchimedesForce(pState);
	m_pForce[1] = new CLaunchRodForce(pState);
	m_pForce[2] = new CAtmosphericPressureForce(pState);
	m_pForce[3] = new CWindResistanceForce(pState);
	m_pForce[4] = new CGravitationForce(pState);
	m_pForce[5] = new CWaterForce(pState);
	m_pForce[6] = new CParachuteForce(pState);
}

CForces::~CForces()
{
	for (int n=0; n<_forces; n++)
		delete m_pForce[n];
}

double CForces::SimulateForces(double dt)
{
	double dForceSum = 0.0;
	for (int n=0; n<_forces; n++)
	{
		m_dForce[n] = m_pForce[n]->Simulate(dt);
		dForceSum += m_dForce[n];
	}
	m_pState->Acceleration = dForceSum / m_pState->CalcTotalMass();
	m_pState->Velocity += m_pState->Acceleration * dt;
	m_pState->Height += m_pState->Velocity * dt;
	return dForceSum;
}


void CreateStateObject(CState *pState, CBottleRocketsApp *pApp)
{
	// get direct pointers to the pages we will need - this will save time
	CEnvironmentSpecsPage *pEnv = pApp->pEnvironmentSpecsPage;
	CLaunchSpecsPage *pLaunch = pApp->pLaunchSpecsPage;
	CRocketSpecsPage *pRocket = pApp->pRocketSpecsPage;
	CSimulationSettingsPage *pSim = pApp->pSimulationSettingsPage;
	
	// get all the data out of fields
	pEnv->UpdateData(TRUE);
	pLaunch->UpdateData(TRUE);
	pRocket->UpdateData(TRUE);
	pSim->UpdateData(TRUE);

	// launcher / environment variables
	pState->LaunchRodDiameter = ConvSL(pEnv->m_dLRDiameter,
		pEnv->m_LRDiameterUnits.GetCurSel(), SL_mm);
	pState->LaunchRodHeight = ConvSL(pEnv->m_dLRHeight,
		pEnv->m_LRHeightUnits.GetCurSel(), SL_mm);
	pState->AtmosphericPressure = ConvP(pEnv->m_dAtmPressure,
		pEnv->m_AtmPressureUnits.GetCurSel(), P_atm);
	pState->AtmosphericTemperature = ConvT(pEnv->m_dAtmTemperature,
		pEnv->m_AtmTemperatureUnits.GetCurSel(), T_degc);
	pState->Gravity = ConvA(pEnv->m_dGravity,
		pEnv->m_GravityUnits.GetCurSel(), A_mps2);
	pState->Updraft = ConvV(pEnv->m_dUpdraft,
		pEnv->m_GravityUnits.GetCurSel(), V_mps);

	// rocket variables
	pState->BottleDiameter = ConvSL(pRocket->m_dDiameter,
		pRocket->m_DiameterUnits.GetCurSel(), SL_mm);
	pState->BottleVolume = ConvVM(pRocket->m_dBottleVolume,
		pRocket->m_BottleVolumeUnits.GetCurSel(), VM_l);
	pState->DragCoefficient = pRocket->m_dDragCoefficient;
	pState->EmptyMass = ConvM(pRocket->m_dEmptyMass,
		pRocket->m_EmptyMassUnits.GetCurSel(), M_kg);

	// parachute variables
	pState->ParachuteKamikaze = pRocket->m_bParKamikaze;
	pState->ParachuteCount = pRocket->m_nParNumber;
	pState->ParachuteDeployTime = pRocket->m_dParDeployTime;
	pState->ParachuteDragCoefficient = pRocket->m_dParDragCoefficient;
	pState->ParachuteSurfaceArea = pRocket->m_nParCircular == 0 ? PI *
		pow(ConvSL(pRocket->m_dParDiameter,
		pRocket->m_ParDiameterUnits.GetCurSel(), SL_cm) / 200.0, 2) :
		ConvSA(pRocket->m_dParSurfaceArea,
		pRocket->m_ParSurfaceAreaUnits.GetCurSel(), SA_m2);
	
	// launch variables
	if ( pLaunch->m_WaterVolumeUnits.GetCurSel() == VM_count )
		// units are already in %
		pState->WaterLevel = pLaunch->m_dWaterVolume / 100.0;
	else
		// convert some other units into %
		pState->WaterLevel = ConvVM(pLaunch->m_dWaterVolume,
			pLaunch->m_WaterVolumeUnits.GetCurSel(), VM_l) /
			pState->BottleVolume;
	pState->InitialWaterLevel = pState->WaterLevel;
	pState->InitialAirPressure = ConvP(pLaunch->m_dInitialPressure,
		pLaunch->m_InitialPressureUnits.GetCurSel(), P_pa);
	pState->AirTemperature = ConvT(pLaunch->m_dAirTemperature,
		pLaunch->m_AirTemperatureUnits.GetCurSel(), T_degc);
	pState->Height = ConvLL(pLaunch->m_dHeight,
		pLaunch->m_HeightUnits.GetCurSel(), LL_m);
	pState->Velocity = ConvV(pLaunch->m_dVelocity,
		pLaunch->m_VelocityUnits.GetCurSel(), V_mps);
	pState->WaterDensity = ConvD(pLaunch->m_dWaterDensity,
		pLaunch->m_WaterDensityUnits.GetCurSel(), D_kgpm3);

	pState->SetInitialAirVolume();
}

int Simulate(double *dX, double *dY, double *dZ, int *nType, BOOL *bGraph,
			 CBottleRocketsApp *pApp)
{
	int nResult = 0;

	// get commonly used pointers
	CSimulationSettingsPage *pSim = pApp->pSimulationSettingsPage;
	CSimulatePage *pPage = pApp->pSimulatePage;

	// set button states
	pPage->m_Simulate.EnableWindow(FALSE);
	pPage->m_Cancel.EnableWindow(TRUE);

	// calculate dt
	int pos = pSim->m_PrecisionSlider.GetPos();
	double dt = pow(10, -pos);

	switch (pSim->m_nSimulationType)
	{
	case 0:
		// Launch Analysis
		{
			double time = 0.0;
			int nLast = 0;
			int nInterval = int(0.01 / dt);
			// adjust the nInterval variable for clearer graphs
			double dCurVal, dMin = 0.0, dMax = 0.0;

			CState state;
			CreateStateObject(&state, pApp);
			CForces forces(&state);

			// summary or graph?
			*bGraph = pSim->m_nLAResults == 1;
			*nType = data_minmax;	// graph type (used by grapher)

			do
			{
				time += dt;
				forces.SimulateForces(dt);
				
				if (!*bGraph)
				{
					// record summary data
				}
				else
				{
					// record graph data
					switch (pSim->m_LAPlot.GetCurSel())
					{
					case FV_Accel:
						dCurVal = ConvA(state.Acceleration, A_mps2,
							pSim->m_LAPlotUnits.GetCurSel());
						break;
					case FV_AirP:
						dCurVal = ConvP(state.CalcAirPressure(), P_pa,
							pSim->m_LAPlotUnits.GetCurSel());
						break;
					case FV_AirVol:
						if (pSim->m_LAPlotUnits.GetCurSel() == VM_count)
							// convert to %
							dCurVal = state.CalcAirVolume() * 100.0 /
								state.BottleVolume;
						else
							// convert to another unit
							dCurVal = ConvV(state.CalcAirVolume(), VM_l,
								pSim->m_LAPlotUnits.GetCurSel());
						break;
					case FV_Height:
						dCurVal = ConvLL(state.Height, LL_m,
							pSim->m_LAPlotUnits.GetCurSel());
						break;
					case FV_LRF:
						dCurVal = forces.m_dForce[force_launchrod];
						break;
					case FV_Mass:
						dCurVal = ConvM(state.CalcTotalMass(), M_kg,
							pSim->m_LAPlotUnits.GetCurSel());
						break;
					case FV_Thrust:
						dCurVal = forces.m_dForce[force_water];
						break;
					case FV_Velocity:
						dCurVal = ConvV(state.Velocity, V_mps,
							pSim->m_LAPlotUnits.GetCurSel());
						break;
					case FV_WL:
						if (pSim->m_LAPlotUnits.GetCurSel() == VM_count)
							// convert to %
							dCurVal = state.WaterLevel * 100.0;
						else
							// convert to another unit
							dCurVal = ConvV(state.WaterLevel *
								state.BottleVolume, VM_l,
								pSim->m_LAPlotUnits.GetCurSel());
						break;
					case FV_WR:
						dCurVal = forces.m_dForce[force_windresistance];
						break;
					}

					// update minimums / maximums for this time interval
					if (dCurVal > dMax || nLast == 0)
						dMax = dCurVal;
					if (dCurVal < dMin || nLast == 0)
						dMin = dCurVal;

					if (nLast++ >= nInterval)
					{
						// time to record the values again
						nLast = 0;
						dX[nResult] = dMin;
						dY[nResult] = dMax;
						dZ[nResult] = time;
						nResult++;
					}
				}
				
				// update progress bar
				pPage->m_Progress.SetPos(int(time * 10) % 100);
				
				// check for cancel
				pPage->PumpMessages();
				if (pPage->m_bCancelled)
					goto abort;
			} while (state.Height > 0.0 && nResult < _max - 1);

			if (bGraph)
			{
				// record one final set of values
				dX[nResult] = dMin;
				dY[nResult] = dMax;
				dZ[nResult] = time;
				nResult++;
			}
		}
		break;

	case 1:
	case 2:
		// 1-Variable Optimization
		// and 2-Variable Optimization
		{
			double dXVal, dYVal;	// x- and y-axis values (used for loop)
			double dXMin, dXMax;	// min and max x-axis values
			double dYMin, dYMax;	// min and max y-axis values (ignored by
									//   1-Variable Optimization
			double dXValReal, dYValReal, dXMinReal, dXMaxReal, dYMinReal,
				dYMaxReal;			// same vars - without unit conversions
			int nXVar, nYVar;		// stores the variable type (used for
									//   calculations)
			int x, y;	// for loops
			int xtests = _testsperaxis;	// the const is used for 1- and 2-var
			int ytests = pSim->m_nSimulationType == 1 ? 1 : _testsperaxis;
				// this variable varies between 1- and 2-var optimizations
			int nDependentVar, nDependentUnits;	// used for converting output

			// a fake state object - we only use this as a cheap way to get
			//   the variables we need for initialization
			CState fakestate;
			CreateStateObject(&fakestate, pApp);

			// get min/max values
			if (pSim->m_nSimulationType == 1)
			{
				// 1-var sim
				nDependentVar = pSim->m_1VDependent.GetCurSel();
				nDependentUnits = pSim->m_1VDependentUnits.GetCurSel();
				nXVar = pSim->m_1VIndependent.GetCurSel();
				nYVar = -1;		// this one isn't used for 1-var sims
				dXMinReal = pSim->m_d1VIndependentMin;
				dXMaxReal = pSim->m_d1VIndependentMax;
				dYMinReal = 0.0;
				dYMaxReal = 0.0;
				dXMin = ConvUnits(dXMinReal, nXVar,
					pSim->m_1VIndependentUnits.GetCurSel(), &fakestate);
				dXMax = ConvUnits(dXMaxReal, nXVar,
					pSim->m_1VIndependentUnits.GetCurSel(), &fakestate);
				dYMin = 0.0;	// these values are ignored - use something
				dYMax = 0.0;	// that computes fast

				*nType = data_points;	// throw this in here - it has to be
										// done sometime
			}
			else
			{
				// 2-var sim
				nDependentVar = pSim->m_2VDependent.GetCurSel();
				nDependentUnits = pSim->m_2VDependentUnits.GetCurSel();
				nXVar = pSim->m_2VXAxis.GetCurSel();
				nYVar = pSim->m_2VYAxis.GetCurSel();
				dXMinReal = pSim->m_d2VXAxisMin;
				dXMaxReal = pSim->m_d2VXAxisMax;
				dYMinReal = pSim->m_d2VYAxisMin;
				dYMaxReal = pSim->m_d2VYAxisMax;
				dXMin = ConvUnits(dXMinReal, nXVar,
					pSim->m_2VXAxisUnits.GetCurSel(), &fakestate);
				dXMax = ConvUnits(dXMaxReal, nXVar,
					pSim->m_2VXAxisUnits.GetCurSel(), &fakestate);
				dYMin = ConvUnits(dYMinReal, nYVar,
					pSim->m_2VYAxisUnits.GetCurSel(), &fakestate);
				dYMax = ConvUnits(dYMaxReal, nYVar,
					pSim->m_2VYAxisUnits.GetCurSel(), &fakestate);

				*nType = data_2var;		// again, this is random but it works
			}
			
			for(x=0; x<xtests; x++)
			{
				// calculate x-value
				dXVal = (x * (dXMax - dXMin) / (xtests - 1)) + dXMin;
				dXValReal = (x * (dXMaxReal - dXMinReal) / (xtests -
					1)) + dXMinReal;

				for(y=0; y<ytests; y++)
				{
					// calculate y-value
					dYVal = (y * (dYMax - dYMin) / (ytests - 1)) + dYMin;
					dYValReal = (y * (dYMaxReal - dYMinReal) / (ytests -
						1)) + dYMinReal;

					// create the state object and initialize
					CState state;
					CreateStateObject(&state, pApp);
					SetVariableToState(dXVal, nXVar, &state);
					if (nYVar != -1)
						SetVariableToState(dYVal, nYVar, &state);

					// simulate!
					double time = 0.0;
					double dValue = 0.0;
					CForces forces(&state);
					do
					{
						// simulate
						time += dt;
						forces.SimulateForces(dt);

						// record the variable we are interested in
						RecordVariable(&dValue, nDependentVar,
							nDependentUnits, &state, time);

						// check for cancel
						pPage->PumpMessages();
						if (pPage->m_bCancelled)
							goto abort;
					} while (state.Height > 0.0);

					// we have the result of the test stored in dValue - now
					//   record it
					// Reminder: x- and y-values must be in original units
					dX[nResult] = dXValReal;
					dY[nResult] = dYValReal;
					dZ[nResult] = dValue;
					nResult++;

					// update progress bar
					pPage->m_Progress.SetPos(int((x * ytests + y + 1) *
						100.0 / (xtests * ytests)));
				}
			}
		}
		break;
	}

abort:
	// reset states
	pPage->m_Simulate.EnableWindow(TRUE);
	pPage->m_Cancel.EnableWindow(FALSE);
	pPage->m_Progress.SetPos(0);

	return nResult;
}


double ConvUnits(double dValue, int nVariable, int nSrcUnits, CState *pState)
{
	switch (nVariable)
	{
	case IV_Diameter:
	case IV_LRDiameter:
	case IV_LRHeight:
		// small length -> mm
		return ConvSL(dValue, nSrcUnits, SL_mm);
	
	case IV_ParDiameter:
		// small length -> cm
		return ConvSL(dValue, nSrcUnits, SL_cm);

	case IV_InitHeight:
		// large length -> m
		return ConvLL(dValue, nSrcUnits, LL_m);

	case IV_InitVelocity:
		// velocity -> m/s
		return ConvV(dValue, nSrcUnits, V_mps);

	case IV_AirT:
	case IV_AtmT:
		// temperature -> deg C
		return ConvT(dValue, nSrcUnits, T_degc);

	case IV_BottleVol:
		// volume -> L
		return ConvVM(dValue, nSrcUnits, VM_l);

	case IV_InitWL:
		if (nSrcUnits == VM_count)
			return dValue / 100.0;	// % -> ratio
		else
			return ConvVM(dValue, nSrcUnits, VM_l) / pState->BottleVolume;
				// volume -> ratio
		// NOTE: this may be a cause for error if Bottle Volume is ever
		//   optimized at the same time as Water Level

	case IV_EmptyMass:
		// mass -> kg
		return ConvM(dValue, nSrcUnits, M_kg);
	
	case IV_AtmP:
		// pressure -> atm
		return ConvP(dValue, nSrcUnits, P_atm);

	case IV_InitAirP:
		// pressure -> Pascals
		return ConvP(dValue, nSrcUnits, P_pa);

	case IV_ParSA:
		// surface area -> m^2
		return ConvSA(dValue, nSrcUnits, SA_m2);

	case IV_WaterD:
		// density -> kg/m^3
		return ConvD(dValue, nSrcUnits, D_kgpm3);

	case IV_Gravity:
		// acceleration -> m/s^2
		return ConvA(dValue, nSrcUnits, A_mps2);

	case IV_Cd:
	case IV_ParDeployTime:
	case IV_ParCd:
	case IV_ParNumber:
		// same units
		return dValue;
	}

	return dValue;
}

void SetVariableToState(double dValue, int nVariable, CState *pState)
{
	switch (nVariable)
	{
	case IV_AirT:
		pState->AirTemperature = dValue;
		pState->SetInitialAirVolume();
		break;
	case IV_AtmP:
		pState->AtmosphericPressure = dValue;
		break;
	case IV_AtmT:
		pState->AtmosphericTemperature = dValue;
		break;
	case IV_BottleVol:
		pState->BottleVolume = dValue;
		break;
	case IV_Cd:
		pState->DragCoefficient = dValue;
		break;
	case IV_Diameter:
		pState->BottleDiameter = dValue;
		break;
	case IV_EmptyMass:
		pState->EmptyMass = dValue;
		break;
	case IV_InitAirP:
		pState->InitialAirPressure = dValue;
		pState->SetInitialAirVolume();
		break;
	case IV_InitHeight:
		pState->Height = dValue;
		break;
	case IV_InitVelocity:
		pState->Velocity = dValue;
		break;
	case IV_InitWL:
		pState->WaterLevel = dValue;
		pState->SetInitialAirVolume();
		break;
	case IV_LRDiameter:
		pState->LaunchRodDiameter = dValue;
		break;
	case IV_LRHeight:
		pState->LaunchRodHeight = dValue;
		break;
	case IV_ParCd:
		pState->ParachuteDragCoefficient = dValue;
		break;
	case IV_ParDeployTime:
		pState->ParachuteDeployTime = dValue;
		break;
	case IV_ParDiameter:
		pState->ParachuteSurfaceArea = PI * pow(dValue / 200, 2);
		break;
	case IV_ParNumber:
		pState->ParachuteCount = (int)dValue;
		break;
	case IV_ParSA:
		pState->ParachuteSurfaceArea = dValue;
		break;
	case IV_Gravity:
		pState->Gravity = dValue;
		break;
	case IV_WaterD:
		pState->WaterDensity = dValue;
		break;
	}
}

void RecordVariable(double *dValue, int nVariable, int nVariableUnits,
					CState *pState, double dTime)
{
	switch (nVariable)
	{
	case DV_ApogeeTime:
		if (pState->Velocity <= 0.0 && *dValue == 0.0)
			*dValue = dTime;
		break;
	case DV_BurnoutTime:
		if (pState->Acceleration <= 0.0 && *dValue == 0.0)
			*dValue = dTime;
		break;
	case DV_ImpactVelocity:
		// see comment on DV_TimeAloft
		*dValue = ConvV(pState->Velocity, V_mps, nVariableUnits);
		break;
	case DV_MaxAccel:
		{
			double dAccel = ConvA(pState->Acceleration, A_mps2,
				nVariableUnits);
			if (dAccel > *dValue)
				*dValue = dAccel;
		}
		break;
	case DV_MaxHeight:
		{
			double dHeight = ConvLL(pState->Height, LL_m, nVariableUnits);
			if (dHeight > *dValue)
				*dValue = dHeight;
		}
		break;
	case DV_MaxVelocity:
		{
			double dVelocity = ConvV(pState->Velocity, V_mps,
				nVariableUnits);
			if (pState->Velocity > *dValue)
				*dValue = pState->Velocity;
		}
		break;
	case DV_TimeAloft:
		// simply update the time every time this function is called - the
		//   the last time will be when the rocket hits the ground
		*dValue = dTime;
		break;
	}
}

