/////////////////////////////////////////////////////////////////////////////
//                                                                         //
//  Columbus High School Bottle Rocket Simulator                           //
//  Copyright (C) 1999-2000 Leo C. Singleton IV                            //
//    Internet: http://www.leosolutions.com/                               //
//    E-Mail: leo@leosolutions.com                                         //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

// forces.cpp
// defines and calculates forces related to rocket

#include "stdafx.h"
#include <math.h>
#include "forces.h"

CForce::CForce(CState *pNewState)
{
	pState = pNewState;
}

CForce::~CForce()
{
}


CArchimedesForce::CArchimedesForce(CState *pNewState) : CForce(pNewState)
{
}

CArchimedesForce::~CArchimedesForce()
{
}

double CArchimedesForce::Simulate(double dt)
{
	if ( pState->WaterLevel > 0.0 )
		return pState->CalcAtmosphericAirDensity() *
			pState->CalcBottleVolumeMinusRod() * -pState->Gravity / 1000;
	else
		return 0.0;
}


CLaunchRodForce::CLaunchRodForce(CState *pNewState) : CForce(pNewState)
{
}

CLaunchRodForce::~CLaunchRodForce()
{
}

double CLaunchRodForce::Simulate(double dt)
{
	if ( pState->CalcRodInBottle() > 0.0 )
		return pState->CalcRodPerpendicularSA() * pState->CalcAirPressure();
	else
		return 0;
}

CAtmosphericPressureForce::CAtmosphericPressureForce(CState *pNewState)
	: CForce(pNewState)
{
}

CAtmosphericPressureForce::~CAtmosphericPressureForce()
{
}

double CAtmosphericPressureForce::Simulate(double dt)
{
	if ( pState->CalcRodInBottle() > 0.0 )
		return pState->CalcRodPerpendicularSA() *
			-pState->AtmosphericPressure * Atm2Pascals;
	else
		return 0.0;
}

CWindResistanceForce::CWindResistanceForce(CState *pNewState)
	: CForce(pNewState)
{
}

CWindResistanceForce::~CWindResistanceForce()
{
}

double CWindResistanceForce::Simulate(double dt)
{
	double wr = pState->CalcAtmosphericAirDensity() *
		pState->CalcRocketPerpendicularSA() * pow(pState->Velocity - pState->Updraft, 2) *
		pState->DragCoefficient / 2;

	if ( pState->Velocity - pState->Updraft > 0.0 )
		return -wr;
	else
		return wr;
}

CGravitationForce::CGravitationForce(CState *pNewState) : CForce(pNewState)
{
}

CGravitationForce::~CGravitationForce()
{
}

double CGravitationForce::Simulate(double dt)
{
	return pState->CalcTotalMass() * pState->Gravity;
}

CWaterForce::CWaterForce(CState *pNewState) : CForce(pNewState)
{
}

CWaterForce::~CWaterForce()
{
}

double CWaterForce::Simulate(double dt)
{
	// if no water is in bottle, this force does not need to be simulated
	if ( pState->WaterLevel == 0.0 )
		return 0.0;

	// also, if rod is in bottle, this force is not simulated
	if ( pState->CalcRodInBottle() > 0.0 )
		return 0.0;

	// calculate ratio of surface area of water to nozzle
	double SA1 = pow(pState->LaunchRodDiameter / 2, 2);	// in mm^2
	double SA2 = pow(pState->BottleDiameter / 2, 2);	// in mm^2

	// calculate velocity of water leaving bottle
/*	double WaterVelocity = sqrt((2 * (pState->CalcAirPressure() -
		(pState->AtmosphericPressure * Atm2Pascals) - (pState->WaterDensity *
		(pState->Acceleration - pState->Gravity) *
		pState->CalcWaterHeight()))) / (pState->WaterDensity * (1 - pow(SA1 /
		SA2, 2))));
*/
	double WaterVelocity = sqrt((2 * (pState->CalcAirPressure() -
		(pState->AtmosphericPressure * Atm2Pascals) + (pState->WaterDensity *
		(pState->Acceleration - pState->Gravity) *
		pState->CalcWaterHeight()))) / (pState->WaterDensity * (1 - pow(SA1 /
		SA2, 2))));


	// calculate water leaving bottle (in L)
	double VolumeLeaving = WaterVelocity * dt * SA1 / 1000;

	// calculate change in water level
	double LevelChange = VolumeLeaving / pState->BottleVolume;

	// subtract water leaving from bottle
	pState->WaterLevel -= LevelChange;
	if ( pState->WaterLevel < 0.0 )
		pState->WaterLevel = 0.0;

	// calculate force caused by water
	return WaterVelocity * VolumeLeaving * pState->WaterDensity / (dt *
		1000);
}

CParachuteForce::CParachuteForce(CState *pNewState) : CForce(pNewState)
{
	m_dDeployCounter = 0.0;
}

CParachuteForce::~CParachuteForce()
{
}

double CParachuteForce::Simulate(double dt)
{
	double pd = 0.0;	// percent deployed (0.0 - 1.0)
	
	if (pState->ParachuteKamikaze)
		return 0.0;	// kamikaze rockets do not use parachutes
	
	if (pState->ParachuteDeployTime <= 0.0)
	{
		// bug fix for Deploy Time = 0
		if (pState->Velocity < 0.0)
			pd = 1.0;
		else
			return 0.0;
	}
	else
	{
		if (m_dDeployCounter < pState->ParachuteDeployTime && pState->Velocity <
			0.0)
		{
			// parachute has not fully deployed, but increment deploy counter
			m_dDeployCounter += dt;
			pd = pow(m_dDeployCounter / pState->ParachuteDeployTime, 10);
		}
		else if (m_dDeployCounter >= pState->ParachuteDeployTime)
		{
			// parachute has fully deployed
			pd = 1.0;
		}
	}

	// calculate parachute force
	double pf = pState->CalcAtmosphericAirDensity() *
		pState->ParachuteSurfaceArea * pState->ParachuteCount * pd *
		pow(-pState->Velocity + pState->Updraft, 2) * pState->ParachuteDragCoefficient / 2;

	if (-pState->Velocity + pState->Updraft > 0)
		return pf;
	else
		return -pf;
}

